from __future__ import absolute_import
from .technique import PositionVariantTechnique
import cv2
import numpy as np
from scipy.ndimage.interpolation import map_coordinates
from scipy.ndimage.filters import gaussian_filter

class elasticTransformAugmentationTechnique(PositionVariantTechnique):

    def __init__(self,parameters):
        PositionVariantTechnique.__init__(self, parameters)
        if 'alpha' in list(parameters.keys()):
            self.alpha = float(parameters["alpha"])
        else:
            self.alpha = 5
        if 'sigma' in list(parameters.keys()):
            self.sigma = float(parameters["sigma"])
        else:
            self.sigma = 0.05

    def __elastic_transform(self,image):
        """Elastic deformation of images as described in [Simard2003]_.
        .. [Simard2003] Simard, Steinkraus and Platt, "Best Practices for
           Convolutional Neural Networks applied to Visual Document Analysis", in
           Proc. of the International Conference on Document Analysis and
           Recognition, 2003.
        """
        random_state = np.random.RandomState(None)
        if len(image.shape) == 3:
            (b,g,r) = cv2.split(image)

            shape = b.shape

            dx = gaussian_filter((random_state.rand(*shape) * 2 - 1), self.sigma, mode="constant", cval=0) * self.alpha
            dy = gaussian_filter((random_state.rand(*shape) * 2 - 1), self.sigma, mode="constant", cval=0) * self.alpha

            x, y = np.meshgrid(np.arange(shape[0]), np.arange(shape[1]), indexing='ij')
            indices = np.reshape(x + dx, (-1, 1)), np.reshape(y + dy, (-1, 1))

            return cv2.merge([map_coordinates(b, indices, order=1).reshape(shape),
                              map_coordinates(g, indices, order=1).reshape(shape),
                              map_coordinates(r, indices, order=1).reshape(shape)])
        else:
            shape = image.shape

            dx = gaussian_filter((random_state.rand(*shape) * 2 - 1), self.sigma, mode="constant", cval=0) * self.alpha
            dy = gaussian_filter((random_state.rand(*shape) * 2 - 1), self.sigma, mode="constant", cval=0) * self.alpha

            x, y = np.meshgrid(np.arange(shape[0]), np.arange(shape[1]), indexing='ij')
            indices = np.reshape(x + dx, (-1, 1)), np.reshape(y + dy, (-1, 1))

            return map_coordinates(image, indices, order=1).reshape(shape)

    def apply(self, image):
        translation = self.__elastic_transform(image)
        return translation

# technique = elasticTransformAugmentationTechnique(2,0.5)
# image = cv2.imread("LPR1.jpg")
# cv2.imshow("t",technique.applyForClassification(image))
# cv2.waitKey(0)



from __future__ import absolute_import
from __future__ import division
from past.utils import old_div
from .technique import PositionVariantTechnique
import cv2

class cropAugmentationTechnique(PositionVariantTechnique):

    # percentage is a value between 0 and 1
    # startFrom indicates the starting point of the cropping,
    # the possible values are TOPLEFT, TOPRIGHT, BOTTOMLEFT,
    # BOTTOMRIGHT, and CENTER
    def __init__(self,parameters):
        PositionVariantTechnique.__init__(self, parameters)
        if 'percentage' in list(parameters.keys()):
            self.percentage = float(parameters["percentage"])
        else:
            self.percentage = 0.9
        if 'startFrom' in list(parameters.keys()):
            self.startFrom = parameters["startFrom"]
        else:
            self.startFrom = 'TOPLEFT'

        if (self.percentage <0 or self.percentage >1):
            raise NameError("Invalid value for cropping")
        if self.startFrom not in ['TOPLEFT', 'TOPRIGHT', 'BOTTOMLEFT', 'BOTTOMRIGHT', 'CENTER']:
            raise NameError("Invalid value for cropping")


    def apply(self, image):
        (h,w) = image.shape[:2]
        newW = int(w*self.percentage)
        newH = int(h * self.percentage)
        if self.startFrom == 'TOPLEFT':
            crop = image[0:newH,0:newW]
        if self.startFrom == 'BOTTOMLEFT':
            crop = image[h-newH:h, 0:newW]
        if self.startFrom == 'TOPRIGHT':
            crop = image[0:newH,w-newW:w]
        if self.startFrom == 'BOTTOMRIGHT':
            crop = image[h-newH:h, w - newW:w]
        if self.startFrom == 'CENTER':
            crop = image[int(old_div(h,2) - old_div(newH,2)):int(old_div(h,2) + old_div(newH,2)),
                   int(old_div(w,2) - old_div(newW,2)):int(old_div(w,2) + old_div(newW,2))]
        return crop


# Example
# technique = cropAugmentationTechnique(0.5,'TOPRIGHT')
# image = cv2.imread("LPR1.jpg")
# cv2.imshow("original",image)
# cv2.imshow("new",technique.apply(image))
# cv2.waitKey(0)


from __future__ import absolute_import
from .technique import PositionInvariantTechnique
import cv2
import numpy as np

class sharpenAugmentationTechnique(PositionInvariantTechnique):

    # Valid values for kernel are 3,5,7,9, and 11
    def __init__(self,parameters):
        PositionInvariantTechnique.__init__(self, parameters)

    def apply(self, image):
        kernel = np.array([[0,-1,0],
                           [-1, 5, -1],
                           [0, -1, 0]])

        sharpen = cv2.filter2D(image, -1, kernel)
        return sharpen


# technique = sharpenAugmentationTechnique()
# image = cv2.imread("LPR1.jpg")
# cv2.imshow("resized",technique.apply(image))
# cv2.waitKey(0)


from __future__ import absolute_import
from .folderKerasLinearClassificationAugmentor import FolderKerasLinearClassificationAugmentor
from .folderKerasSemanticSegmentationAugmentor import FolderKerasSemanticSegmentationAugmentor
from .folderLinearClassificationAugmentor import FolderLinearClassificationAugmentor
from .folderLinearSemanticSegmentationAugmentor import FolderLinearSemanticSegmentationAugmentor
from .hdf5LinearClassificationAugmentor import HDF5LinearClassificationAugmentor
from .hdf5LinearLocalizationAugmentor import  HDF5LinearLocalizationAugmentor
from .hdf5LinearSegmentationAugmentor import HDF5LinearSegmentationAugmentor
from .hdf5PowerClassificationAugmentor import HDF5PowerClassificationAugmentor
from .hdf5PowerSegmentationAugmentor import HDF5PowerSegmentationAugmentor
from .pascalVOCLinearDetectionAugmentor import PascalVOCLinearDetectionAugmentor
from .pascalVOCLinearLocalizationAugmentor import PascalVOCLinearLocalizationAugmentor
from .yoloLinearDetectionAugmentor import yoloLinearDetectionAugmentor
from .cocoLinearInstanceSegmentationAugmentor import COCOLinearInstanceSegmentationAugmentor
from .folderVideoLinearClassificationAugmentor import FolderVideoLinearClassificationAugmentor
from .youtubeVideoLinearDetectionAugmentor import YoutubeVideoLinearDetectionAugmentor
from .folderStackLinearSemanticSegmentationAugmentor import FolderStackLinearSemanticSegmentationAugmentor
from .jsonLinearInstanceSegmentationAugmentor import JSONLinearInstanceSegmentationAugmentor

# List of augmentors organized by problem, then annotation mode, then output mode, and
# finally generation
Augmentors = {
    "classification" : {
        "folders": {
            "folders": {
                "linear" : FolderLinearClassificationAugmentor,
                },
            "hdf5": {
                "linear" : HDF5LinearClassificationAugmentor,
                "power" : HDF5PowerClassificationAugmentor,
                },
            "keras": {
                "linear" : FolderKerasLinearClassificationAugmentor,
                },
            },
        },

    "localization" : {
        "pascalvoc": {
            "pascalvoc" : {
                "linear" :PascalVOCLinearLocalizationAugmentor
                },
            "hdf5": {
                "linear": HDF5LinearLocalizationAugmentor,
            },


            },
        },

    "detection" : {
        "pascalvoc": {
            "pascalvoc" : {
                "linear" : PascalVOCLinearDetectionAugmentor
                },

            },
        "yolo": {
            "yolo": {
                "linear": yoloLinearDetectionAugmentor
            },

        },
        },

    "semantic_segmentation" : {
        "folders": {
            "folders": {
                "linear": FolderLinearSemanticSegmentationAugmentor,
            },
            "hdf5": {
                "linear": HDF5LinearSegmentationAugmentor,
                "power": HDF5PowerSegmentationAugmentor,
            },
            "keras": {
                "linear": FolderKerasSemanticSegmentationAugmentor,
            },
        },
    },
    "instance_segmentation" : {
        "coco": {
            "coco": {
                "linear": COCOLinearInstanceSegmentationAugmentor,
            },
        },
        "json": {
            "json": {
                "linear": JSONLinearInstanceSegmentationAugmentor,
            },
        },

    },

   "stackclassification" : {
        "videofolders": {
            "videofolders": {
                "linear" : FolderVideoLinearClassificationAugmentor,
                },
            },
        },
   "stackdetection" : {
        "youtubevideo": {
            "youtubevideo": {
                "linear" : YoutubeVideoLinearDetectionAugmentor,
                },
            },
        },
   "stacksegmentation" : {
        "tiff": {
            "tiff": {
                "linear" : FolderStackLinearSemanticSegmentationAugmentor,
                },
            },
        },
}



from __future__ import absolute_import
from builtins import str
from builtins import object
from .iaugmentor import IAugmentor
from imutils import paths
import os
import cv2
from sklearn.externals.joblib import Parallel, delayed
import pandas as pd
import numpy as np


# We need to define this function outside to work in parallel.
def readAndGenerateVideo(outputPath, transformers, i_and_videoPath, df):
    (i, videoPath) = i_and_videoPath
    name = videoPath.split(os.path.sep)[-1]
    ext = name[name.rfind("."):]
    name = name[0:name.rfind(".")]

    boxes = np.array(df[df[0] == name].iloc[:, 1:])
    boxesConverted = []

    vidcap = cv2.VideoCapture(videoPath)
    fps = vidcap.get(cv2.CAP_PROP_FPS)
    listImages = []
    success, image = vidcap.read()
    (hI, wI) = image.shape[:2]
    for box in boxes:
        if box[1] != -1.00:
            category = box[1]
            x = int(box[5] * wI)
            y = int(box[7] * hI)
            h = int((box[8] - y) * hI)
            w = int((box[6] - x) * wI)
            boxesConverted.append((category, (x, y, w, h)))
    listImages.append(image)
    while success:
        success, image = vidcap.read()
        listImages.append(image)
    vidcap.release()
    boxesOutput = []

    for (j, transformer) in enumerate(transformers):
        newListImages, newboxes = transformer.transform(listImages, boxesConverted)

        out = cv2.VideoWriter(outputPath + str(i) + "_" + str(j) + "_" + name + ext,
                              cv2.VideoWriter_fourcc('M', 'J', 'P', 'G'), fps, (wI, hI))

        for image in newListImages:
            out.write(image)

        out.release()
        for box0, box1 in zip(boxes, newboxes):
            (c, (x, y, wb, hb)) = box1
            boxesOutput.append([str(i) + "_" + str(j) + "_" + name, box0[0], c,
                                box0[2], box0[3], box0[4], x / wI, (x + wb) / wI, y / hI, (y + hb) / hI])

    return boxesOutput


# This class serves to generate videos for a classification
# problem where all the videos are organized by folders
# distributed by labels. Example:
# - Folder
# |- cats
#    |- video1.mp4
#    |- video2.mp4
#    |- ...
# |- dogs
#    |- video1.mp4
#    |- video2.mp4
#    |- ...
class YoutubeVideoLinearDetectionAugmentor(IAugmentor):

    def __init__(self, inputPath, parameters):
        IAugmentor.__init__(self)
        self.inputPath = inputPath
        # output path represents the folder where the images will be stored
        if parameters["outputPath"]:
            self.outputPath = parameters["outputPath"]
        else:
            raise ValueError("You should provide an output path in the parameters")
        if parameters["csv"]:
            self.csv = parameters["csv"]
        else:
            raise ValueError("You should provide a csv name containing the annotations")

    def readImagesAndAnnotations(self):
        self.imagePaths = list(paths.list_files(self.inputPath, validExts=(".avi", ".mp4")))

    def createOutputDirs(self):
        if not (os._exists(self.outputPath)):
            os.makedirs(self.outputPath)

    def applyAugmentation(self):
        self.readImagesAndAnnotations()
        self.createOutputDirs()
        df = pd.read_csv(self.inputPath + self.csv, header=None)

        # [readAndGenerateVideo(self.outputPath,self.generators,x) for x in enumerate(self.imagePaths)]
        boxesOutput = Parallel(n_jobs=-1)(
            delayed(readAndGenerateVideo)(self.outputPath, self.transformers, x, df) for x in
            enumerate(self.imagePaths))
        boxesOutput = [item for sublist in boxesOutput for item in sublist]

        df = pd.DataFrame(boxesOutput)
        df.to_csv(self.outputPath + self.csv, header=False, index=False)



from __future__ import absolute_import
from .transformer import Transformer
from sklearn.externals.joblib import Parallel, delayed


class TransformerForImageStackClassification(Transformer):

    def __init__(self,technique,dictLabels=None):
        Transformer.__init__(self,technique,dictLabels)


    def transform(self, listImages,label):
        newlistImages = Parallel(n_jobs=-1)(
            delayed(self.technique.apply)(image) for image in listImages if image is not None)
        return [newlistImages,self.transformLabel(label)]





from __future__ import absolute_import
from .technique import PositionInvariantTechnique
import cv2
import numpy as np

class invertAugmentationTechnique(PositionInvariantTechnique):


    def __init__(self,parameters):
        PositionInvariantTechnique.__init__(self, parameters)


    def apply(self, image):

        image1 = cv2.bitwise_not(image)
        return image1


from __future__ import absolute_import
from .transformer import Transformer
from sklearn.externals.joblib import Parallel, delayed
from ..techniques.technique import PositionVariantTechnique
from .detection import detectBoxes



class TransformerForImageStackDetection(Transformer):

    def __init__(self,technique,dictLabels=None):
        Transformer.__init__(self,technique,dictLabels)


    def transform(self, listImages, boxes):
        imageShape = listImages[0].shape[:2]

        if (isinstance(self.technique, PositionVariantTechnique)):
            newBoxes = detectBoxes(imageShape, boxes, self.technique)
            newBoxes = [(self.transformLabel(box[0]),box[1]) for box in newBoxes]
        else:
            newBoxes= [(self.transformLabel(box[0]),box[1]) for box in boxes]

        newlistImages = Parallel(n_jobs=-1)(
            delayed(self.technique.apply)(image) for image in listImages if image is not None)
        return [newlistImages,newBoxes]


from __future__ import absolute_import
from .transformer import Transformer
from sklearn.externals.joblib import Parallel, delayed
from ..techniques.technique import PositionVariantTechnique



class TransformerForImageStackSegmentation(Transformer):

    def __init__(self,technique,dictLabels=None):
        Transformer.__init__(self,technique,dictLabels)


    def transform(self, listImages, listMasks):


        newlistImages = Parallel(n_jobs=-1)(
            delayed(self.technique.apply)(image) for image in listImages if image is not None)
        if (isinstance(self.technique, PositionVariantTechnique)):
            newListMasks = Parallel(n_jobs=-1)(
            delayed(self.technique.apply)(image) for image in listMasks if image is not None)
            return [newlistImages,newListMasks]
        else:
            return [newlistImages, listMasks]


from __future__ import absolute_import
from .augmentors.augmentorFactory import createAugmentor
from .transformers.transformerFactory import transformerGenerator
from .techniques.techniqueFactory import createTechnique
import argparse
from .utils.conf import Conf


def main(conf):
	#ap = argparse.ArgumentParser()
	#ap.add_argument("-c", "--conf", required=True, help="path to configuration file")
	#args = vars(ap.parse_args())


	conf = Conf(conf)

	# First, we read the parameters
	problem = conf["problem"]
	annotationMode = conf["annotation_mode"]
	outputMode = conf["output_mode"]
	generationMode = conf["generation_mode"]
	inputPath = conf["input_path"]
	parameters = conf["parameters"]
	augmentationTechniques = conf["augmentation_techniques"]

	# Second, we create the augmentor
	augmentor = createAugmentor(problem,annotationMode,outputMode,generationMode,inputPath,
		                    parameters)

	# Third, we load the techniques and add them to the augmentor
	techniques = [createTechnique(technique,parameters) for (technique,parameters) in
		      augmentationTechniques]

	transformer = transformerGenerator(problem)

	for technique in techniques:
	    augmentor.addTransformer(transformer(technique))

	# Finally, we apply the augmentation
	augmentor.applyAugmentation()




from __future__ import absolute_import
from builtins import zip
from builtins import object
from mahotas.demos import image_path

from .iaugmentor import IAugmentor
from imutils import paths
import os
import cv2
from sklearn.externals.joblib import Parallel, delayed
import random
import numpy as np
from .utils.aspectawarepreprocessor import AspectAwarePreprocessor

def readAndGenerateImageSegmentation(image,label,transformers):
    newimage = image
    newlabel = label
    for (j, transformer) in enumerate(transformers):
        if (random.randint(0, 100) > 50):
            newimage,newlabel = transformer.transform(newimage,newlabel)

    return (newimage,newlabel)

# This class serves to generate images for a semantic segmentation
# problem where all the images are organized in two folders called
# images and labels. Both folders must contain the same number of images
# and with the same names (but maybe different extensions). Example
# - Folder
# |- images
#    |- image1.jpg
#    |- image2.jpg
#    |- ...
# |- labels
#    |- image1.tiff
#    |- image2.tiff
#    |- ...
# where Folder/labels/image1.tiff is the annotation of the image Folder/images/image1.jpg.
# Hence, both images must have the same size.
class FolderKerasSemanticSegmentationAugmentor(IAugmentor):

    def __init__(self,inputPath,parameters):
        IAugmentor.__init__(self)
        self.inputPath = inputPath
        self.imagesPath = inputPath+"images/"
        self.labelsPath = inputPath + "labels/"
        # output path represents the folder where the images will be stored
        if parameters["labelsExtension"]:
            self.labelsExtension=parameters["labelsExtension"]
        else:
            self.labelsExtension = ".tiff"
        if parameters["width"]:
            self.width = parameters["width"]
        else:
            self.width = 64
        if parameters["height"]:
            self.height = parameters["height"]
        else:
            self.height = 64

        if parameters["batchSize"]:
            self.batchSize = parameters["batchSize"]
        else:
            self.batchSize = 32
        self.readImagesAndAnnotations()


    def readImagesAndAnnotations(self):
        self.imagePaths = list(paths.list_files(self.imagesPath,validExts=(".jpg", ".jpeg", ".png", ".bmp",".tiff",".tif")))
        self.labelPaths = list(paths.list_files(self.labelsPath,validExts=(".jpg", ".jpeg", ".png", ".bmp",".tiff",".tif")))
        self.numImages = len(self.imagePaths)
        if (len(self.imagePaths)!=len(self.labelPaths)):
            raise Exception("The number of files is different in the folder of images and in the folder of labels")



    def applyAugmentation(self,passes=np.inf):
        epochs = 0
        aap = AspectAwarePreprocessor(self.width, self.height)
        while epochs < passes:

            for i in np.arange(0, self.numImages, self.batchSize):
                imagPaths = self.imagePaths[i:i+self.batchSize]
                labPaths = self.labelPaths[i:i+self.batchSize]
                images = [aap.preprocess(cv2.imread(imagePath)) for imagePath in imagPaths]
                labels = [aap.preprocess(cv2.imread(labelPath)) for labelPath in labPaths]
                images_labels = [readAndGenerateImageSegmentation(image,label,self.transformers) for (image,label) in zip(images,labels)]
                images = [i[0] for i in images_labels]
                labels = [i[1] for i in images_labels]
                yield (images,labels)


            epochs += 1





from __future__ import absolute_import
from builtins import str
from builtins import object
import numpy as np

from .iaugmentor import IAugmentor
from .utils.readCOCOJSON import readCOCOJSON
import json
import cv2
from sklearn.externals.joblib import Parallel, delayed
import imutils
from imutils import paths
import os


def readAndGenerateInstanceSegmentation(outputPath, transformers, imagePath):
    image = cv2.imread(imagePath)
    name = imagePath.split(os.path.sep)[-1]
    labelPath = '/'.join(imagePath.split(os.path.sep)[:-1]) + "/" + name[0:name.rfind(".")] + ".txt"
    maskLabels = []
    (h,w)=image.shape[:2]
    with open(labelPath) as f:
        data = json.load(f)

    annotations = data['annotations']


    for annotation in annotations:
        mask = np.zeros((w, h), dtype="uint8")
        key = list(annotation)[0]

        if key == 'rectangle':
            x = annotation['rectangle']['x']
            y = annotation['rectangle']['y']
            width = annotation['rectangle']['width']
            height = annotation['rectangle']['height']
            label = annotation['rectangle']['label']
            cv2.rectangle(mask,(x,y),(x+width,y+height),255,-1)
            maskLabels.append((mask, label))

        if key == 'circle':
            x = annotation['circle']['x']
            y = annotation['circle']['y']
            diameter = annotation['circle']['radius']
            label = annotation['circle']['label']
            cv2.circle(mask,(int(x+diameter/2),int(y+diameter/2)),int(diameter/2),255,-1)
            maskLabels.append((mask, label))

        if key == 'polygon':
            xpoints = annotation['polygon']['xpoints']
            ypoints = annotation['polygon']['ypoints']
            label = annotation['polygon']['label']
            pts = np.array([[x,y] for x,y in zip(xpoints,ypoints)], np.int32)
            pts = pts.reshape((-1, 1, 2))
            cv2.fillPoly(mask, [pts], True, 255)

            maskLabels.append((mask, label))

    for (j, transformer) in enumerate(transformers):
        (newimage, newmasklabels) = transformer.transform(image, maskLabels)

        cv2.imwrite(outputPath + str(j) + "_" + name, newimage)
        newSegmentations = []
        for (mask, label) in newmasklabels:
            cnts = cv2.findContours(mask.copy(), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
            cnts = cnts[0] if imutils.is_cv2() else cnts[1]
            segmentation = [[x[0][0], x[0][1]] for x in cnts[0]]
            # Closing the polygon
            segmentation.append(segmentation[0])

            newSegmentations.append((label,  segmentation))
        data = {}
        data['name'] = name
        data['width'] = w
        data['height'] = h
        data['annotations'] = []
        for (l,segmentation) in newSegmentations:
            xpoints = [int(x[0]) for x in segmentation]
            ypoints = [int(x[1]) for x in segmentation]
            data['annotations'].append({'polygon':{'ypoints':ypoints,
                                                   'xpoints':xpoints,
                                                   'label':label}})


        with open(outputPath + str(j) + "_" + name[0:name.rfind(".")]+ ".txt", 'w') as outfile:
            json.dump(data, outfile)


# This class serves to generate images for an instance segmentation
# problem where all the images are organized in a folder together with
# a special purpose json (a json file for each image).

class JSONLinearInstanceSegmentationAugmentor(IAugmentor):

    def __init__(self, inputPath, parameters):
        IAugmentor.__init__(self)
        self.inputPath = inputPath
        # output path represents the folder where the images will be stored
        if parameters["outputPath"]:
            self.outputPath = parameters["outputPath"]
        else:
            raise ValueError("You should provide an output path in the parameters")


    def readImagesAndAnnotations(self):
        self.imagePaths = list(
            paths.list_files(self.inputPath, validExts=(".jpg", ".jpeg", ".png", ".bmp", ".tiff", ".tif")))
        self.labelPaths = list(paths.list_files(self.inputPath, validExts=(".txt")))
        if (len(self.imagePaths) != len(self.labelPaths)):
            raise Exception("The number of images is different to the number of annotations")

    def applyAugmentation(self):
        self.readImagesAndAnnotations()

        Parallel(n_jobs=-1)(delayed(readAndGenerateInstanceSegmentation)
                                             (self.outputPath, self.transformers, imagePath)
                                             for imagePath in self.imagePaths)





from __future__ import absolute_import
from builtins import object
from .iaugmentor import IAugmentor
from sklearn.preprocessing import LabelEncoder
from imutils import paths
import os
import cv2
import progressbar
from .utils.aspectawarepreprocessor import AspectAwarePreprocessor
from .utils.hdf5datasetwriter import HDF5DatasetWriterSegmentation

# This class serves to generate images for a classification
# problem where all the images are organized by folders
# distributed by labels. Example:
# - Folder
# |- cats
#    |- image1.jpg
#    |- image2.jpg
#    |- ...
# |- dogs
#    |- image1.jpg
#    |- image2.jpg
#    |- ...
class HDF5LinearSegmentationAugmentor(IAugmentor):

    # All images must have same width and height
    def __init__(self,inputPath,parameters):
        IAugmentor.__init__(self)
        self.inputPath = inputPath
        self.imagesPath = inputPath + "images/"
        self.labelsPath = inputPath + "labels/"
        # output path represents the folder where the images will be stored
        if parameters["outputPath"]:
            self.outputPath = parameters["outputPath"]
        else:
            raise ValueError("You should provide an output path in the parameters")

        if parameters["labelsExtension"]:
            self.labelsExtension = parameters["labelsExtension"]
        else:
            self.labelsExtension = ".tiff"
        if parameters["width"]:
            self.width = parameters["width"]
        else:
            raise ValueError("You should provide a width in the parameters")
        if parameters["height"]:
            self.width = parameters["height"]
        else:
            raise ValueError("You should provide a height in the parameters")
        self.aw = AspectAwarePreprocessor(self.width,self.height)



    def readImagesAndAnnotations(self):
        self.imagePaths = list(
            paths.list_files(self.imagesPath, validExts=(".jpg", ".jpeg", ".png", ".bmp", ".tiff", ".tif")))
        self.labelPaths = list(
            paths.list_files(self.labelsPath, validExts=(".jpg", ".jpeg", ".png", ".bmp", ".tiff", ".tif")))
        if (len(self.imagePaths) != len(self.labelPaths)):
            raise Exception("The number of files is different in the folder of images and in the folder of labels")

    def applyAugmentation(self):
        self.readImagesAndAnnotations()
        widgets = ["Processing images: ", progressbar.Percentage(), " ",
                   progressbar.Bar(), " ", progressbar.ETA()]

        pbar = progressbar.ProgressBar(maxval=len(self.imagePaths),
                                   widgets=widgets).start()
        writer = HDF5DatasetWriterSegmentation((len(self.imagePaths)*len(self.transformers),self.width,self.height,3),
                                   self.outputPath)

        for i_and_imagePath in enumerate(self.imagePaths):
            (i, imagePath) = i_and_imagePath
            image = cv2.imread(imagePath)
            image = self.aw.preprocess(image)
            name = imagePath.split(os.path.sep)[-1]
            labelPath = '/'.join(imagePath.split(os.path.sep)[:-2]) + "/labels/" + name[
                                                                                   0:name.rfind(".")] + self.labelsExtension
            label = cv2.imread(labelPath)
            label = self.aw.preprocess(label)
            for (j, transformer) in enumerate(self.transformers):
                (newimage, newlabel) = transformer.transform(image, label)
                writer.add([newimage],[newlabel])
            pbar.update(i)
        writer.close()
        pbar.finish()







from __future__ import absolute_import
from .technique import PositionInvariantTechnique
import cv2
import numpy as np

class gausianBlurringAugmentationTechnique(PositionInvariantTechnique):

    # Valid values for kernel are 3,5,7,9, and 11
    def __init__(self,parameters):
        PositionInvariantTechnique.__init__(self, parameters)
        if 'kernel' in list(parameters.keys()):
            self.kernel = int(parameters["kernel"])
        else:
            self.kernel = 3
        if (not (self.kernel in [3,5,7,9,11])):
            raise NameError("Invalid value for kernel")

    def apply(self, image):
        blurred = cv2.GaussianBlur(image, (self.kernel, self.kernel),0)
        return blurred


from __future__ import absolute_import
from builtins import str
from builtins import object
from mahotas.demos import image_path

from .iaugmentor import IAugmentor
from imutils import paths
import os
import cv2
from sklearn.externals.joblib import Parallel, delayed


def readAndGenerateImageSegmentation(outputPath, transformers, labelextension, i_and_imagePath):
    (i, imagePath) = i_and_imagePath
    image = cv2.imread(imagePath)
    name = imagePath.split(os.path.sep)[-1]
    labelPath = '/'.join(imagePath.split(os.path.sep)[:-2]) + "/labels/" + name[
                                                                           0:name.rfind(".")] + labelextension
    label = cv2.imread(labelPath)


    for (j, transformer) in enumerate(transformers):
        (newimage,newlabel) = transformer.transform(image,label)

        cv2.imwrite(outputPath +  "images/" + str(i) + "_" + str(j) + "_" + name,
                    newimage)
        cv2.imwrite(outputPath + "labels/" + str(i) + "_" + str(j) + "_" + name[0:name.rfind(".")]+labelextension,
                    newlabel)

# This class serves to generate images for a semantic segmentation
# problem where all the images are organized in two folders called
# images and labels. Both folders must contain the same number of images
# and with the same names (but maybe different extensions). Example
# - Folder
# |- images
#    |- image1.jpg
#    |- image2.jpg
#    |- ...
# |- labels
#    |- image1.tiff
#    |- image2.tiff
#    |- ...
# where Folder/labels/image1.tiff is the annotation of the image Folder/images/image1.jpg.
# Hence, both images must have the same size.
class FolderLinearSemanticSegmentationAugmentor(IAugmentor):

    def __init__(self,inputPath,parameters):
        IAugmentor.__init__(self)
        self.inputPath = inputPath
        self.imagesPath = inputPath+"/images/"
        self.labelsPath = inputPath + "/labels/"
        # output path represents the folder where the images will be stored
        if parameters["outputPath"]:
            self.outputPath = parameters["outputPath"]
        else:
            raise ValueError("You should provide an output path in the parameters")
        if parameters["labelsExtension"]:
            self.labelsExtension = parameters["labelsExtension"]
        else:
            self.labelsExtension = ".tiff"
        if not os.path.exists(self.outputPath + "images/"):
            os.makedirs(self.outputPath + "images/")
        if not os.path.exists(self.outputPath + "labels/"):
            os.makedirs(self.outputPath + "labels/")

    def readImagesAndAnnotations(self):

        self.imagePaths = list(paths.list_files(self.imagesPath,validExts=(".jpg", ".jpeg", ".png", ".bmp",".tiff",".tif")))
        self.labelPaths = list(paths.list_files(self.labelsPath,validExts=(".jpg", ".jpeg", ".png", ".bmp",".tiff",".tif")))
        if (len(self.imagePaths)!=len(self.labelPaths)):
            raise Exception("The number of files is different in the folder of images and in the folder of labels")



    def applyAugmentation(self):
        self.readImagesAndAnnotations()
        Parallel(n_jobs=-1)(delayed(readAndGenerateImageSegmentation)
                            (self.outputPath,self.transformers,self.labelsExtension,x)
                            for x in enumerate(self.imagePaths))



# Example
# augmentor = FolderLinearSemanticSegmentationAugmentor(
#     "/home/joheras/pythonprojects/ssai-cnn/maps/mass_buildings/test/",
#     "/home/joheras/pythonprojects/ssai-cnn/maps/mass_buildings/test/dataset-generated/",
#     ".tif"
# )

# from techniques.averageBlurringAugmentationTechnique import averageBlurringAugmentationTechnique
# from techniques.bilateralBlurringAugmentationTechnique import bilateralBlurringAugmentationTechnique
# from techniques.gaussianNoiseAugmentationTechnique import gaussianNoiseAugmentationTechnique
# from techniques.rotateAugmentationTechnique import rotateAugmentationTechnique
# from techniques.flipAugmentationTechnique import flipAugmentationTechnique
# from techniques.noneAugmentationTechnique import noneAugmentationTechnique
# from generator import Generator
# import time
# augmentor.addGenerator(Generator(noneAugmentationTechnique()))
# augmentor.addGenerator(Generator(averageBlurringAugmentationTechnique()))
# augmentor.addGenerator(Generator(bilateralBlurringAugmentationTechnique()))
# augmentor.addGenerator(Generator(gaussianNoiseAugmentationTechnique()))
# augmentor.addGenerator(Generator(rotateAugmentationTechnique()))
# augmentor.addGenerator(Generator(flipAugmentationTechnique()))
# start = time.time()
# augmentor.applyAugmentation()
# end = time.time()
# print(end - start)




from __future__ import absolute_import
from builtins import zip
from builtins import object
from .iaugmentor import IAugmentor
from sklearn.preprocessing import LabelEncoder
from imutils import paths
import os
import cv2
from sklearn.externals.joblib import Parallel, delayed
from .utils.aspectawarepreprocessor import AspectAwarePreprocessor
from .utils.hdf5datasetwriter import HDF5DatasetWriterClassification
import progressbar

# This class serves to generate images for a classification
# problem where all the images are organized by folders
# distributed by labels. Example:
# - Folder
# |- cats
#    |- image1.jpg
#    |- image2.jpg
#    |- ...
# |- dogs
#    |- image1.jpg
#    |- image2.jpg
#    |- ...
class HDF5PowerClassificationAugmentor(IAugmentor):

    # All images must have same width and height
    def __init__(self,inputPath,parameters):
        IAugmentor.__init__(self)
        self.inputPath = inputPath
        # output path represents the h5py file where dataset will be stored
        if parameters["outputPath"]:
            self.outputPath = parameters["outputPath"]
        else:
            raise ValueError("You should provide an output path in the parameters")

        if parameters["width"]:
            self.width = parameters["width"]
        else:
            raise ValueError("You should provide a width in the parameters")
        if parameters["height"]:
            self.width = parameters["height"]
        else:
            raise ValueError("You should provide a height in the parameters")

        self.aw = AspectAwarePreprocessor(self.width, self.height)


    def readImagesAndAnnotations(self):
        self.imagePaths = list(paths.list_files(self.inputPath,validExts=(".jpg", ".jpeg", ".png", ".bmp",".tiff",".tif")))


    def applyAugmentation(self):
        self.readImagesAndAnnotations()
        le = LabelEncoder()
        labels = [p.split(os.path.sep)[-2] for p in self.imagePaths]
        labels = le.fit_transform(labels)
        writer = HDF5DatasetWriterClassification((len(self.imagePaths)*(2**(len(self.transformers)-1)),self.width,self.height,3),
                                   self.outputPath)
        # We need to define this function outside to work in parallel.
        writer.storeClassLabels(le.classes_)
        widgets = ["Processing images: ", progressbar.Percentage(), " ",
                   progressbar.Bar(), " ", progressbar.ETA()]
        pbar = progressbar.ProgressBar(maxval=len(self.imagePaths),
                                       widgets=widgets).start()
        for i_and_imagePath in enumerate(zip(self.imagePaths,labels)):
            (i, (imagePath,label)) = i_and_imagePath
            image = cv2.imread(imagePath)
            image = self.aw.preprocess(image)

            images = [image]

            for (j, transformer) in enumerate(self.transformers):
                newimages = []
                for (k,im) in enumerate(images):
                    newimage,newlabel = transformer.transform(im,label)
                    newimage = self.aw.preprocess(newimage)
                    writer.add([newimage], [newlabel])
                    newimages.append(newimage)
                images = newimages

            pbar.update(i)
        writer.close()
        pbar.finish()

# # Example
# augmentor = HDF5PowerClassificationAugmentor(
#     "/home/joheras/datasets/cats_and_dogs_small/train/",
#     "/home/joheras/datasets/cats_and_dogs_small/database.hdf5",224,224
# )

# from techniques.averageBlurringAugmentationTechnique import averageBlurringAugmentationTechnique
# from techniques.bilateralBlurringAugmentationTechnique import bilateralBlurringAugmentationTechnique
# from techniques.gaussianNoiseAugmentationTechnique import gaussianNoiseAugmentationTechnique
# from techniques.rotateAugmentationTechnique import rotateAugmentationTechnique
# from techniques.flipAugmentationTechnique import flipAugmentationTechnique
# from techniques.noneAugmentationTechnique import noneAugmentationTechnique
# from generator import Generator
# import time
# augmentor.addGenerator(Generator(noneAugmentationTechnique()))
# augmentor.addGenerator(Generator(averageBlurringAugmentationTechnique()))
# augmentor.addGenerator(Generator(bilateralBlurringAugmentationTechnique()))
# augmentor.addGenerator(Generator(gaussianNoiseAugmentationTechnique()))
# augmentor.addGenerator(Generator(rotateAugmentationTechnique()))
# augmentor.addGenerator(Generator(flipAugmentationTechnique()))
# start = time.time()
# augmentor.applyAugmentation()
# end = time.time()
# print(end - start)



from __future__ import absolute_import
from .technique import PositionInvariantTechnique
import cv2
import numpy as np

class saltAndPepperNoiseAugmentationTechnique(PositionInvariantTechnique):


    def __init__(self,parameters):
        PositionInvariantTechnique.__init__(self, parameters)
        if 'low' in list(parameters.keys()):
            self.low = float(parameters["low"])
        else:
            self.low = 0
        if 'up' in list(parameters.keys()):
            self.up = float(parameters["up"])
        else:
            self.up = 25

    def apply(self, image):
        im = np.zeros(image.shape, np.uint8)
        if(len(image.shape)!=3):
            l = self.low
            u = self.up
        else:
            l = (self.low,self.low,self.low)
            u = (self.up,self.up,self.up)
        cv2.randu(im, l,u)
        image_noise = cv2.add(image, im)
        return image_noise


# image = cv2.imread("LPR1.jpg")
# print (image.shape)
# im = np.zeros(image.shape, np.uint8)
# m = (50,50,50)
# s = (10,10,10)
# cv2.randu(im, (0,0,0),(50,50,50))
# image_noise=cv2.add(image, im)
# cv2.imshow("original",image)
# cv2.imshow("noise",image_noise)
# cv2.waitKey(0)


from __future__ import absolute_import
from .techniques.techniqueFactory import createTechnique
import cv2
import numpy as np

# In this file, it is shown how the different techniques can be applied
# individually to an image.


# Loading the image
img = cv2.imread("sample_images/cat1.jpg")

t = createTechnique("cropSize", {"x" : 0,"y":0,"width":20,"height":20})
img1 = t.apply(img)
cv2.imshow("cropSize",np.hstack([img,img1]))
cv2.waitKey(0)
#
# t = createTechnique("bilateral_blurring", {"diameter" : 11, "sigmaColor": 21, "sigmaSpace":7})
# img1 = t.apply(img)
# cv2.imwrite("bilateral_blurring.jpg",np.hstack([img,img1]))
#
# t = createTechnique("blurring", {"ksize" : 5})
# img1 = t.apply(img)
# cv2.imwrite("blurring.jpg",np.hstack([img,img1]))
#
# t = createTechnique("change_to_hsv",{})
# img1 = t.apply(img)
# cv2.imwrite("change_to_hsv.jpg",np.hstack([img,img1]))
#
# t = createTechnique("change_to_lab",{})
# img1 = t.apply(img)
# cv2.imwrite("change_to_lab.jpg",np.hstack([img,img1]))
#
# t = createTechnique("crop",{"percentage":0.8,"startFrom": "TOPLEFT"})
# img1 = t.apply(img)
# cv2.imwrite("crop.jpg",img1)
#
# t = createTechnique("dropout",{"percentage":0.05})
# img1 = t.apply(img)
# cv2.imwrite("dropout.jpg",np.hstack([img,img1]))
#
# t = createTechnique("elastic",{"alpha":5,"sigma":0.05})
# img1 = t.apply(img)
# cv2.imwrite("elastic.jpg",np.hstack([img,img1]))
#
# t = createTechnique("equalize_histogram",{})
# img1 = t.apply(img)
# cv2.imwrite("equalize.jpg",np.hstack([img,img1]))
#
# t = createTechnique("flip",{"flip":0})
# t1 = createTechnique("flip",{"flip":1})
# t2 = createTechnique("flip",{"flip":-1})
# img1 = t.apply(img)
# img2 = t1.apply(img)
# img3 = t2.apply(img)
# cv2.imwrite("flip.jpg",np.hstack([img,img1,img2,img3]))
#
#
# t = createTechnique("gamma",{"gamma":1.5})
# img1 = t.apply(img)
# cv2.imwrite("gamma.jpg",np.hstack([img,img1]))
#
# t = createTechnique("gaussian_blur", {"kernel" : 5})
# img1 = t.apply(img)
# cv2.imwrite("gaussian_blurring.jpg",np.hstack([img,img1]))
#
# t = createTechnique("gaussian_noise", {"mean" : 0,"sigma":10})
# img1 = t.apply(img)
# cv2.imwrite("gaussian_noise.jpg",np.hstack([img,img1]))
#
# t = createTechnique("invert",{})
# img1 = t.apply(img)
# cv2.imwrite("invert.jpg",np.hstack([img,img1]))
#
# t = createTechnique("median_blur", {"kernel" : 5})
# img1 = t.apply(img)
# cv2.imwrite("median_blurring.jpg",np.hstack([img,img1]))
#
#
# t = createTechnique("none",{})
# img1 = t.apply(img)
# cv2.imwrite("none.jpg",np.hstack([img,img1]))
#
#
# t = createTechnique("raise_blue", {"power" : 0.9})
# img1 = t.apply(img)
# cv2.imwrite("raise_blue.jpg",np.hstack([img,img1]))
#
# t = createTechnique("raise_green", {"power" : 0.9})
# img1 = t.apply(img)
# cv2.imwrite("raise_green.jpg",np.hstack([img,img1]))
#
# t = createTechnique("raise_red", {"power" : 0.9})
# img1 = t.apply(img)
# cv2.imwrite("raise_red.jpg",np.hstack([img,img1]))
#
# t = createTechnique("raise_hue", {"power" : 0.9})
# img1 = t.apply(img)
# cv2.imwrite("raise_hue.jpg",np.hstack([img,img1]))
#
# t = createTechnique("raise_saturation", {"power" : 0.9})
# img1 = t.apply(img)
# cv2.imwrite("raise_saturation.jpg",np.hstack([img,img1]))
#
# t = createTechnique("raise_value", {"power" : 0.9})
# img1 = t.apply(img)
# cv2.imwrite("raise_value.jpg",np.hstack([img,img1]))
#
#
# t = createTechnique("resize", {"percentage" : 0.9,"method":"INTER_NEAREST"})
# img1 = t.apply(img)
# cv2.imwrite("resize.jpg",img1)
#
# t = createTechnique("rotate", {"angle" : 90})
# img1 = t.apply(img)
# cv2.imwrite("rotate.jpg",img1)
#
# t = createTechnique("salt_and_pepper", {"low" : 0,"up":25})
# img1 = t.apply(img)
# cv2.imwrite("salt_and_peper.jpg",np.hstack([img,img1]))
#
#
# t = createTechnique("sharpen", {})
# img1 = t.apply(img)
# cv2.imwrite("sharpen.jpg",np.hstack([img,img1]))
#
# t = createTechnique("shift_channel", {"shift":0.2})
# img1 = t.apply(img)
# cv2.imwrite("sharpen.jpg",np.hstack([img,img1]))
#
# t = createTechnique("shearing", {"a":0.5})
# img1 = t.apply(img)
# cv2.imwrite("shearing.jpg",np.hstack([img,img1]))
#
# t = createTechnique("translation", {"x":10,"y":10})
# img1 = t.apply(img)
# cv2.imwrite("translation.jpg",np.hstack([img,img1]))

